@@ -42,6 +42,7 @@
 #include "obex.h"
 #include "dbus.h"
 #include "logging.h"
+#include "btio.h"
 
 static const gchar *opp_record = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>	\
 <record>									\
@@ -231,11 +232,12 @@ struct agent {
 static struct agent *agent = NULL;
 
 struct pending_request {
+	DBusPendingCall *call;
 	struct server *server;
-	gchar address[18];
 	gchar *adapter_path;
+	char address[18];
 	guint watch;
-	gint nsk;
+	GIOChannel *io;
 };
 
 struct adapter_any {
@@ -640,7 +642,7 @@ done:
 
 	dbus_message_unref(reply);
 }
-static void find_adapter(const char *pattern,
+static DBusPendingCall *find_adapter(const char *pattern,
 				DBusPendingCallNotifyFunction function,
 				gpointer user_data)
 {
@@ -657,21 +659,29 @@ static void find_adapter(const char *pattern,
 
 	dbus_connection_send_with_reply(system_conn, msg, &call, -1);
 	dbus_pending_call_set_notify(call, function, user_data, NULL);
-	dbus_pending_call_unref(call);
 
 	dbus_message_unref(msg);
+
+	return call;
 }
 
-static gboolean find_adapter_any(gpointer user_data)
+static gboolean find_adapter_any_idle(gpointer user_data)
 {
-	find_adapter("any", find_adapter_any_reply, user_data);
+	DBusPendingCall *call;
+
+	call = find_adapter("any", find_adapter_any_reply, NULL);
+	dbus_pending_call_unref(call);
 
 	return FALSE;
 }
 
 static void name_acquired(DBusConnection *conn, void *user_data)
 {
-	find_adapter_any(NULL);
+	DBusPendingCall *call;
+
+	call = find_adapter("any", find_adapter_any_reply, NULL);
+	dbus_pending_call_unref(call);
+
 	bluetooth_start();
 }
 
@@ -709,7 +719,7 @@ gboolean manager_init(void)
 	listener_id = g_dbus_add_service_watch(system_conn, "org.bluez",
 				name_acquired, name_released, NULL, NULL);
 
-	g_idle_add(find_adapter_any, NULL);
+	g_idle_add(find_adapter_any_idle, NULL);
 
 	return g_dbus_register_interface(connection, OPENOBEX_MANAGER_PATH,
 					OPENOBEX_MANAGER_INTERFACE,
@@ -830,6 +840,16 @@ void unregister_transfer(guint32 id)
 	g_free(path);
 }
 
+static void agent_cancel()
+{
+	DBusMessage *msg;
+
+	msg = dbus_message_new_method_call(agent->bus_name, agent->path,
+					"org.openobex.Agent", "Cancel");
+
+	g_dbus_send_message(connection, msg);
+}
+
 static void agent_reply(DBusPendingCall *call, gpointer user_data)
 {
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
@@ -849,6 +869,10 @@ static void agent_reply(DBusPendingCall *call, gpointer user_data)
 	if (dbus_set_error_from_message(&derr, reply)) {
 		error("Agent replied with an error: %s, %s",
 				derr.name, derr.message);
+
+		if (dbus_error_has_name(&derr, DBUS_ERROR_NO_REPLY))
+			agent_cancel();
+
 		dbus_error_free(&derr);
 		dbus_message_unref(reply);
 		return;
@@ -956,11 +980,7 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 
 	if (!got_reply) {
 		dbus_pending_call_cancel(call);
-		msg = dbus_message_new_method_call(agent->bus_name,
-							agent->path,
-							"org.openobex.Agent",
-							"Cancel");
-		g_dbus_send_message(connection, msg);
+		agent_cancel();
 	}
 
 	dbus_pending_call_unref(call);
@@ -976,31 +996,74 @@ int request_authorization(gint32 cid, int fd, const gchar *filename,
 	return 0;
 }
 
+static void service_cancel(struct pending_request *pending)
+{
+	DBusMessage *msg;
+
+	msg = dbus_message_new_method_call("org.bluez",
+					pending->adapter_path,
+					"org.bluez.Service",
+					"CancelAuthorization");
+
+	g_dbus_send_message(system_conn, msg);
+}
+
+void obex_connect_cb(GIOChannel *io, GError *err, gpointer user_data)
+{
+	struct server *server = user_data;
+
+	if (err) {
+		error("%s", err->message);
+		g_io_channel_shutdown(io, TRUE, NULL);
+		return;
+	}
+
+	if (obex_session_start(io, server) < 0)
+		g_io_channel_shutdown(io, TRUE, NULL);
+}
+
+static void pending_request_free(struct pending_request *pending)
+{
+	if (pending->call)
+		dbus_pending_call_unref(pending->call);
+	g_io_channel_unref(pending->io);
+	g_free(pending->adapter_path);
+	g_free(pending);
+}
+
 static void service_reply(DBusPendingCall *call, gpointer user_data)
 {
 	struct pending_request *pending = user_data;
+	GIOChannel *io = pending->io;
+	struct server *server = pending->server;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	DBusError derr;
+	GError *err = NULL;
 
 	dbus_error_init(&derr);
 	if (dbus_set_error_from_message(&derr, reply)) {
 		error("RequestAuthorization error: %s, %s",
 				derr.name, derr.message);
+
+		if (dbus_error_has_name(&derr, DBUS_ERROR_NO_REPLY))
+			service_cancel(pending);
+
 		dbus_error_free(&derr);
-		close(pending->nsk);
+		g_io_channel_shutdown(io, TRUE, NULL);
 		goto done;
 	}
 
 	debug("RequestAuthorization succeeded");
 
-	if (obex_session_start(pending->nsk, pending->server) < 0)
-		close(pending->nsk);
+	if (!bt_io_accept(io, obex_connect_cb, server, NULL, &err)) {
+		error("%s", err->message);
+		g_error_free(err);
+		g_io_channel_shutdown(io, TRUE, NULL);
+	}
 
 done:
-	if (pending->watch)
-		g_source_remove(pending->watch);
-	g_free(pending->adapter_path);
-	g_free(pending);
+	g_source_remove(pending->watch);
+	pending_request_free(pending);
 	dbus_message_unref(reply);
 }
 
@@ -1008,18 +1071,12 @@ static gboolean service_error(GIOChannel *io, GIOCondition cond,
 			gpointer user_data)
 {
 	struct pending_request *pending = user_data;
-	DBusMessage *msg;
 
-	pending->watch = 0;
+	service_cancel(pending);
 
-	msg = dbus_message_new_method_call("org.bluez",
-					pending->adapter_path,
-					"org.bluez.Service",
-					"CancelAuthorization");
+	dbus_pending_call_cancel(pending->call);
 
-	dbus_connection_send(system_conn, msg, NULL);
-
-	dbus_message_unref(msg);
+	pending_request_free(pending);
 
 	return FALSE;
 }
@@ -1030,9 +1087,7 @@ static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	DBusMessage *msg;
 	DBusPendingCall *pcall;
-	GIOChannel *io;
-	const char *paddr = pending->address;
-	const char *path;
+	const char *path, *paddr = pending->address;
 	DBusError derr;
 
 	dbus_error_init(&derr);
@@ -1068,56 +1123,52 @@ static void find_adapter_reply(DBusPendingCall *call, gpointer user_data)
 
 	debug("RequestAuthorization(%s, %x)", paddr, pending->server->handle);
 
-	if (!dbus_pending_call_set_notify(pcall, service_reply, pending, NULL))
+	if (!dbus_pending_call_set_notify(pcall, service_reply, pending,
+								NULL)) {
+		dbus_pending_call_unref(pcall);
 		goto failed;
+	}
 
-	dbus_pending_call_unref(pcall);
+	dbus_pending_call_unref(pending->call);
+	pending->call = pcall;
 
 	/* Catches errors before authorization response comes */
-	io = g_io_channel_unix_new(pending->nsk);
-	pending->watch = g_io_add_watch_full(io, G_PRIORITY_DEFAULT,
+	pending->watch = g_io_add_watch(pending->io,
 					G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-					service_error, pending, NULL);
-	g_io_channel_unref(io);
+					service_error, pending);
 
 	return;
 
 failed:
-	g_free(pending->adapter_path);
-	close(pending->nsk);
-	g_free(pending);
+	g_io_channel_shutdown(pending->io, TRUE, NULL);
+	pending_request_free(pending);
 }
 
-gint request_service_authorization(struct server *server, gint nsk)
+gint request_service_authorization(struct server *server, GIOChannel *io,
+					const char *address)
 {
-	struct sockaddr_rc src, dst;
-	socklen_t addrlen;
-	gchar address[18];
-	const gchar *paddr = address;
 	struct pending_request *pending;
+	char source[18];
+	GError *err = NULL;
 
 	if (system_conn == NULL || any->path == NULL)
 		return -1;
 
-	memset(&src, 0, sizeof(src));
-	memset(&dst, 0, sizeof(dst));
-	addrlen = sizeof(src);
-
-	if (getsockname(nsk, (struct sockaddr *) &src, &addrlen) < 0)
-		return -1;
-
-	ba2str(&src.rc_bdaddr, address);
-
-	if (getpeername(nsk, (struct sockaddr *) &dst, &addrlen) < 0)
-		return -1;
+	bt_io_get(io, BT_IO_RFCOMM, &err,
+			BT_IO_OPT_SOURCE, source,
+			BT_IO_OPT_INVALID);
+	if (err) {
+		error("%s", err->message);
+		g_error_free(err);
+		return -EINVAL;
+	}
 
 	pending = g_new0(struct pending_request, 1);
 	pending->server = server;
-	pending->nsk = nsk;
+	pending->io = g_io_channel_ref(io);
+	memcpy(pending->address, address, sizeof(pending->address));
 
-	ba2str(&dst.rc_bdaddr, pending->address);
-
-	find_adapter(paddr, find_adapter_reply, pending);
+	pending->call = find_adapter(source, find_adapter_reply, pending);
 
 	return 0;
 }
